/*
 * CS 106B Trailblazer
 * This file declares the graph type used in trailblazer
 *
 * Please do not modify this provided file. Your turned-in files should work
 * with an unmodified version of all provided code files.
 *
 * @author Chris Piech, Keith Schwarz
 * @version 2017/03/09 (updated version for Win17)
 */


#include "RoadGraph.h"
#include "point.h"
#include <math.h>
#include <sstream>

/* Private helper functions only needed in this file. */
namespace {
    /*
     * The number of pixels that separate a and b
     */
    double pointDistance(const Point& a, const Point& b) {
        double dx = a.getX() - b.getX();
        double dy = a.getY() - b.getY();
        return fmax(sqrt(dx * dx + dy * dy), 0) - 1; // TODO: Why is there a -1 here?
    }
}

/* Constructs a new road node with the given name. */
RoadNode::RoadNode(const std::string& name, const Point& location)
    : name(name), myLocation(location) {
    // Everything else handled by default
}

/* Returns the name of the node. */
std::string RoadNode::nodeName() const {
    return name;
}

/* Returns the position of the node on screen. */
Point RoadNode::location() const {
    return myLocation;
}

/* Returns a set of all outgoing edges leaving this node. */
Set<RoadEdge*> RoadNode::outgoingEdges() const {
    return arcs;
}

/* Sets the color of this node in the display. Note that we don't actually remember the node
 * color in this class.
 */
void RoadNode::setColor(Color c) {
    notifyObservers(c);
}

/* Produces a nice, human-readable representation of a road node. */
std::string RoadNode::toString() const {
    std::ostringstream result;
    result << "[" << name << " " << myLocation << "]";
    return result.str();
}

/* Constructs a new edge. */
RoadEdge::RoadEdge(RoadNode* start, RoadNode* finish, double edgeCost)
    : start(start), finish(finish), edgeCost(edgeCost) {
    // Everything else handled by default
}

/* Returns the cost of the edge. */
double RoadEdge::cost() const {
    return edgeCost;
}

/* Returns which node the edge leaves from. */
RoadNode* RoadEdge::from() const {
    return start;
}

/* Returns which node the edges enters. */
RoadNode* RoadEdge::to() const {
    return finish;
}

/* Returns a nice, human-readable representation of the node. */
std::string RoadEdge::toString() const {
    std::ostringstream result;
    result << "[Edge " << start->toString() << " -> " << finish->toString() << " (cost " << edgeCost << ")]";
    return result.str();
}


/*
 * Makes a new RoadGraph based on BasicGraph data
 */
RoadGraph::RoadGraph(Graph<RoadNode, RoadEdge>* data) {
    this->data = data;
}

/*
 * Returns all of the verticies that can be reached by a
 * a direct path from v.
 */
Set<RoadNode *> RoadGraph::neighborsOf(RoadNode* v) const {
    return data->getNeighbors(v);
}

/*
 * Returns a pointer to the edge that connects start to end
 */
RoadEdge* RoadGraph::edgeBetween(RoadNode* start, RoadNode* end) const {
    return data->getArc(start, end);
}

/*
 * The geodesic distance between the two nodes. This is the
 * distance that a crow would travel if it flew from start to
 * end. Units are not standarized, but are consistent within
 * a graph.
 */
double RoadGraph::crowFlyDistanceBetween(RoadNode* start, RoadNode* end) const {
    return pointDistance(start->location(), end->location());
}

/*
 * Returns the maximum speed of any edge on the road graph.
 */
double RoadGraph::maxRoadSpeed() const {
    if(maxRateCached) return maxRate;

    /* Look at every edge in the network and find the one that has the highest travel rate. */
    for(RoadEdge* edge: data->getArcSet()) {
        /* Travel time is equal to the edge cost. */
        double time = edge->cost();

        /* Compute the distance between the two points. */
        Point a = edge->from()->location();
        Point b = edge->to()->location();
        double dist = pointDistance(a, b) - 1; // TODO: Why is there a -1 here?

        if (dist <= 3) continue; // TODO: Why is this check here?

        double rate = dist / time;
        if(rate > maxRate) {
            maxRate = rate;
        }
    }
    maxRateCached = true;
    return maxRate;
}
